"""
test_base_model.py - Unit tests for BaseModel functionality

Copyright(C) 2024 by
Trey V. Wenger; tvwenger@gmail.com
This code is licensed under MIT license (see LICENSE for details)
"""

from contextlib import ExitStack as does_not_raise
import pytest

import numpy as np
import pymc as pm
import arviz.labels as azl

from bayes_spec import BaseModel, SpecData

_RNG = np.random.RandomState(seed=1234)


class ModelA(BaseModel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._cluster_features += ["x", "z"]
        self.var_name_map.update({"x": "X", "y": "Y", "z": "Z"})


class ModelB(ModelA):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def add_priors(self, prior_baseline_coeffs=None):
        super().add_baseline_priors(prior_baseline_coeffs=prior_baseline_coeffs)
        with self.model:
            x = pm.Normal("x", mu=0.0, sigma=1.0, dims="cloud")
            y = pm.Normal("y", mu=0.0, sigma=1.0)
            _ = pm.Deterministic("z", x + y, dims="cloud")


class ModelC(ModelB):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def add_likelihood(self):
        with self.model:
            _ = pm.Normal(
                "observation",
                mu=np.ones_like(self.data["observation"].spectral)
                * self.model["z"].sum(),
                sigma=self.data["observation"].noise,
                observed=self.data["observation"].brightness,
            )


spectral = np.linspace(-5.0, 10.0, 1000)
brightness = 10.0 * _RNG.randn(1000) + 10.0
data = {"observation": SpecData(spectral, brightness, 1.0)}


def test_valid():
    with pytest.raises(TypeError):
        _ = ModelA(data, 1)
    with pytest.raises(TypeError):
        _ = ModelB(data, 1)
    with does_not_raise():
        _ = ModelC(data, 1)


def test_attributes():
    model = ModelC(data, 2, baseline_degree=3, seed=1234, verbose=True)
    model.add_priors()
    with pytest.raises(ValueError):
        model._validate()
    model.add_likelihood()
    assert model.n_clouds == 2
    assert model.baseline_degree == 3
    assert model.seed == 1234
    assert model.verbose is True
    assert model.data == data
    assert model.baseline_freeRVs == ["baseline_observation_norm"]
    assert model.baseline_deterministics == []
    assert model.cloud_freeRVs == ["x"]
    assert model.cloud_deterministics == ["z"]
    assert model.hyper_freeRVs == ["y"]
    assert model.hyper_deterministics == []
    assert model._n_data == 1000
    assert model._n_params == 7
    with pytest.raises(ValueError):
        _ = model._get_unique_solution
    with pytest.raises(ValueError):
        _ = model.unique_solution
    assert isinstance(model.labeller, azl.MapLabeller)
    assert model._validate()


def test_baseline():
    model = ModelC(data, 2, baseline_degree=3, seed=1234, verbose=True)
    model.add_priors()
    model.add_likelihood()

    baseline_params = {
        "baseline_observation_norm": [0.0, 0.0, 0.0, 0.0],
    }
    baseline_model = model.predict_baseline(baseline_params=baseline_params)
    assert len(baseline_model["observation"].eval()) == len(spectral)

    # regression test for #24
    prior_baseline_coeffs = {"observation": [0.0]}
    with pytest.raises(ValueError):
        model.add_priors(prior_baseline_coeffs=prior_baseline_coeffs)



"""
test_spec_data.py - Unit tests for spec_data.py

Copyright(C) 2024 by
Trey V. Wenger; tvwenger@gmail.com
This code is licensed under MIT license (see LICENSE for details)
"""

from contextlib import ExitStack as does_not_raise
import pytest

import numpy as np

from bayes_spec import SpecData

_RNG = np.random.RandomState(seed=1234)


@pytest.mark.parametrize(
    "spectral, brightness, noise, expectation",
    [
        (_RNG.randn(10), _RNG.randn(10), _RNG.randn(10), does_not_raise()),
        (_RNG.randn(10), _RNG.randn(10), 1.0, does_not_raise()),
        (_RNG.randn(10), _RNG.randn(5), 1.0, pytest.raises(ValueError)),
        (_RNG.randn(10), _RNG.randn(10), _RNG.randn(15), pytest.raises(ValueError)),
    ],
)
def test_shape(spectral, brightness, noise, expectation):
    with expectation:
        _ = SpecData(spectral, brightness, noise)


def test_normalize():
    spectral = np.linspace(-5.0, 10.0, 1000)
    brightness = 10.0 * _RNG.randn(1000) + 10.0
    data = SpecData(spectral, brightness, 1.0)
    assert data.spectral_norm.max() == pytest.approx(1.0)
    assert data.spectral_norm.min() == pytest.approx(-1.0)
    assert data.unnormalize_brightness(
        data.normalize_brightness(brightness)
    ) == pytest.approx(brightness)
    assert data.unnormalize_spectral(
        data.normalize_spectral(spectral)
    ) == pytest.approx(spectral)



"""
gauss_model.py
Defines GaussModel, a Gaussian line profile model.

Copyright(C) 2024 by
Trey V. Wenger; tvwenger@gmail.com
This code is licensed under MIT license (see LICENSE for details)
"""

from typing import Iterable, Optional

import numpy as np
import pymc as pm

from bayes_spec import BaseModel
from bayes_spec.utils import gaussian


class GaussModel(BaseModel):
    """
    Definition of a Gaussian line profile model.
    """

    def __init__(self, *args, **kwargs):
        """Initialize a new GaussModel instance

        :param `*args`: Arguments passed to :class:`BaseModel`
        :param `**kwargs`: Keyword arguments passed to :class:`BaseModel`
        """
        # Initialize BaseModel
        super().__init__(*args, **kwargs)

        # Select features used for posterior clustering
        self._cluster_features += [
            "velocity",
            "line_area",
        ]

        # Define TeX representation of each parameter
        self.var_name_map.update(
            {
                "line_area": r"$\int\!T_B\,dV$ (K km s$^{-1}$)",
                "fwhm": r"$\Delta V$ (km s$^{-1}$)",
                "velocity": r"$V_{\rm LSR}$ (km s$^{-1}$)",
                "amplitude": r"$T_B$ (K)",
            }
        )

    def add_priors(
        self,
        prior_line_area: float = 100.0,
        prior_fwhm: float = 25.0,
        prior_velocity: Iterable[float] = [0.0, 25.0],
        prior_baseline_coeffs: Optional[Iterable[float]] = None,
        ordered: bool = False,
    ):
        """Add priors to the model.

        :param prior_line_area: Prior distribution on line area (K km s-1), where
            line_area ~ Gamma(alpha=2.0, beta=1.0/prior_line_area)
            defaults to 100.0
        :type prior_line_area: float, optional
        :param prior_fwhm: Prior distribution on line FWHM (km s-1), where
            fwhm ~ Gamma(alpha=2.0, beta=1.0/prior_fwhm)
            defaults to 25.0
        :type prior_fwhm: float, optional
        :param prior_velocity: Prior distribution on line centroid velocity (km s-1), where
            velocity ~ Normal(mu=prior_velocity[0], sigma=prior_velocity[1]) if :param:ordered is `False`
            velocity(cloud=N) ~ prior_velocity[0] + sum(velocity(cloud<N)) + Gamma(alpha=2.0, beta=1.0/prior_velocity[1]) if :param:ordered is `True`
            defaults to [0.0, 25.0]
        :type prior_velocity: Iterable[float], optional
        :param prior_baseline_coeffs: Width of normal prior distribution on the normalized baseline polynomial
            coefficients. If None, use `[1.0]*(baseline_degree+1)`, defaults to None
        :type prior_baseline_coeff: float, optional
        :param ordered: If True, assume ordered velocities, defaults to False
        :type ordered: bool
        """
        # check inputs
        if not isinstance(prior_line_area, float):
            raise ValueError("prior_line_area must be a number")
        if not isinstance(prior_fwhm, float):
            raise ValueError("prior_fwhm must be a number")
        if not isinstance(prior_velocity, list) or len(prior_velocity) != 2:
            raise ValueError("prior_velocity must be a list of two numbers")
        if prior_baseline_coeffs is not None:
            if not isinstance(prior_baseline_coeffs, list) or len(prior_baseline_coeffs) != self.baseline_degree + 1:
                raise ValueError("prior_baseline_coeffs must be a list of length baseline_degree + 1")

        # add polynomial baseline priors
        if prior_baseline_coeffs is not None:
            prior_baseline_coeffs = {"observation": prior_baseline_coeffs}
        super().add_baseline_priors(prior_baseline_coeffs=prior_baseline_coeffs)

        with self.model:
            # Line area per cloud
            line_area_norm = pm.Gamma("line_area_norm", alpha=2.0, beta=1.0, dims="cloud")
            line_area = pm.Deterministic("line_area", prior_line_area * line_area_norm, dims="cloud")

            # FWHM line width per cloud
            fwhm_norm = pm.Gamma("fwhm_norm", alpha=2.0, beta=1.0, dims="cloud")
            fwhm = pm.Deterministic("fwhm", prior_fwhm * fwhm_norm, dims="cloud")

            # Centroid velocity per cloud
            if ordered:
                velocity_norm = pm.Gamma("velocity_norm", alpha=2.0, beta=1.0, dims="cloud")
                velocity_offset = velocity_norm * prior_velocity[1]
                _ = pm.Deterministic("velocity", prior_velocity[0] + pm.math.cumsum(velocity_offset), dims="cloud")
            else:
                velocity_norm = pm.Normal("velocity_norm", mu=0.0, sigma=1.0, dims="cloud")
                _ = pm.Deterministic("velocity", prior_velocity[0] + prior_velocity[1] * velocity_norm, dims="cloud")

            # Deterministic amplitude per cloud
            _ = pm.Deterministic("amplitude", line_area / fwhm / np.sqrt(np.pi / (4.0 * np.log(2.0))), dims="cloud")

    def predict(self) -> Iterable[float]:
        """Predict observed spectrum from model parameters.

        :return: Predicted spectrum
        :rtype: Iterable[float]
        """
        # Evaluate line profile model per cloud, sum over clouds
        predicted_line = gaussian(
            self.data["observation"].spectral[:, None],
            self.model["amplitude"],
            self.model["velocity"],
            self.model["fwhm"],
        ).sum(axis=1)

        # Add baseline model
        baseline_models = self.predict_baseline()
        predicted = predicted_line + baseline_models["observation"]
        return predicted

    def add_likelihood(self):
        """Add likelihood to the model. Data key must be "observation"."""
        # Predict emission
        predicted = self.predict()

        with self.model:
            # Evaluate likelihood
            _ = pm.Normal(
                "observation",
                mu=predicted,
                sigma=self.data["observation"].noise,
                observed=self.data["observation"].brightness,
            )



"""
nuts.py - customize pymc's NUTS initialization

Copyright(C) 2024 by
Trey V. Wenger; tvwenger@gmail.com
This code is licensed under MIT license (see LICENSE for details)
"""

from typing import Optional

import pymc as pm
from pymc.variational.callbacks import CheckParametersConvergence
from pymc.step_methods.hmc import quadpotential


def init_nuts(
    model: pm.Model,
    init: str = "advi+adapt_diag",
    n_init: int = 100_000,
    chains: int = 4,
    rel_tolerance: float = 0.001,
    abs_tolerance: float = 0.001,
    learning_rate: float = 1e-3,
    start: Optional[dict] = None,
    nuts_kwargs: dict = None,
    seed: int = 1234,
    verbose: bool = False,
) -> tuple[list, pm.NUTS]:
    """_summary_

    :param model: Model to initialize
    :type model: pm.Model
    :param init: Initialization strategy, defaults to "advi+adapt_diag"
    :type init: str, optional
    :param n_init: Number of initialization iterations, defaults to 100_000
    :type n_init: int, optional
    :param chains: Number of independent Markov chains, defaults to 4
    :type chains: int, optional
    :param rel_tolerance: VI relative convergence threshold, defaults to 0.001
    :type rel_tolerance: float, optional
    :param abs_tolerance: VI absolute convergence threshold, defaults to 0.001
    :type abs_tolerance: float, optional
    :param learning_rate: VI learning rate, defaults to 1e-3
    :type learning_rate: float, optional
    :param start: Starting point, defaults to None
    :type start: Optional[dict], optional
    :param nuts_kwargs: Additional keyword arguments passed to :class:`pm.NUTS`, defaults to None
    :type nuts_kwargs: dict, optional
    :param seed: Random seed, defaults to 1234
    :type seed: int, optional
    :param verbose: Verbose output, defaults to False
    :type verbose: bool, optional
    :return: Initial point and step method
    :rtype: tuple[list, pm.NUTS]
    """
    initial_points = None
    step = None
    allowed_init = ["advi+adapt_diag"]
    if init not in allowed_init:
        return initial_points, step

    with model:
        if verbose:
            print(f"Initializing NUTS using custom {init} strategy")
        callbacks = [
            CheckParametersConvergence(tolerance=rel_tolerance, diff="relative"),
            CheckParametersConvergence(tolerance=abs_tolerance, diff="absolute"),
        ]

        if init == "advi+adapt_diag":
            approx = pm.fit(
                random_seed=seed,
                n=n_init,
                method="advi",
                model=model,
                start=start,
                callbacks=callbacks,
                progressbar=verbose,
                obj_optimizer=pm.adagrad_window(learning_rate=learning_rate),
            )
            approx_sample = approx.sample(
                draws=chains,
                random_seed=seed,
                return_inferencedata=False,
            )
            initial_points = [approx_sample[i] for i in range(chains)]
            std_apoint = approx.std.eval()
            cov = std_apoint**2
            mean = approx.mean.get_value()
            n = len(cov)
            potential = quadpotential.QuadPotentialDiagAdapt(n, mean, cov, 50)

        step = pm.NUTS(potential=potential, **nuts_kwargs)

        # Filter deterministics from initial_points
        value_var_names = [var.name for var in model.value_vars]
        initial_points = [
            {k: v for k, v in initial_point.items() if k in value_var_names}
            for initial_point in initial_points
        ]

    return initial_points, step



"""
test_gauss_models.py - Unit tests for GaussLine models and
bayes_spec sampling functions

Copyright(C) 2024 by
Trey V. Wenger; tvwenger@gmail.com
This code is licensed under MIT license (see LICENSE for details)
"""

import numpy as np

from matplotlib.axes import Axes

from bayes_spec import SpecData
from bayes_spec.plots import plot_predictive, plot_pair, plot_traces
from bayes_spec.models import GaussModel

_RNG = np.random.RandomState(seed=1234)


def test_plots():
    # Simulate single-component model
    noise = 1.0
    spectral = np.linspace(-100.0, 100.0, 1000)
    brightness = noise * _RNG.randn(1000)
    data = {"observation": SpecData(spectral, brightness, noise)}
    params = {
        "line_area": [1000.0],
        "fwhm": [25.0],
        "velocity": [10.0],
        "baseline_observation_norm": [0.0],
    }
    model = GaussModel(data, 1, baseline_degree=0, seed=1234, verbose=True)
    model.add_priors()
    model.add_likelihood()
    brightness = model.model["observation"].eval(params)

    # Test plotting functions
    data = {"observation": SpecData(spectral, brightness, noise)}
    model = GaussModel(data, 1, baseline_degree=0, seed=1234, verbose=True)
    model.add_priors()
    model.add_likelihood()
    model.sample(
        chains=2,
        cores=2,
        init_kwargs={"rel_tolerance": 0.01, "abs_tolerance": 0.1, "learning_rate": 1e-2},
    )
    model.solve()
    posterior = model.sample_posterior_predictive()
    assert isinstance(plot_predictive(model.data, posterior.posterior_predictive).ravel()[0], Axes)
    assert isinstance(
        plot_pair(model.trace.solution_0, model.cloud_deterministics, labeller=model.labeller).ravel()[0], Axes
    )
    assert isinstance(plot_traces(model.trace.solution_0, model.cloud_deterministics).ravel()[0], Axes)



"""
utils.py - Utility functions

Copyright(C) 2024 by
Trey V. Wenger; tvwenger@gmail.com
This code is licensed under MIT license (see LICENSE for details)
"""

import pytensor.tensor as pt


def gaussian(x: float, amp: float, center: float, fwhm: float) -> float:
    """Evaluate a Gaussian function

    :param x: Position at which to evaluate
    :type x: float
    :param amp: Gaussian amplitude
    :type amp: float
    :param center: Gaussian centroid
    :type center: float
    :param fwhm: Gaussian full-width at half-maximum
    :type fwhm: float
    :return: Gaussian evaluated at :param:x
    :rtype: float
    """
    return amp * pt.exp(-4.0 * pt.log(2.0) * (x - center) ** 2.0 / fwhm**2.0)



"""
test_utils.py - Unit tests for utils.py

Copyright(C) 2024 by
Trey V. Wenger; tvwenger@gmail.com
This code is licensed under MIT license (see LICENSE for details)
"""

import pytest

from bayes_spec.utils import gaussian


@pytest.mark.parametrize(
    "x, amp, center, fwhm, expectation",
    [
        (0.0, 1.0, 0.0, 1.0, 1.0),
        (0.5, 1.0, 0.0, 1.0, 0.5),
        (0.0, 2.0, 1.0, 2.0, 1.0),
    ],
)
def test_gaussian(x, amp, center, fwhm, expectation):
    assert gaussian(x, amp, center, fwhm).eval() == pytest.approx(expectation)



"""
cluster_posterior.py - Utilities for clustering posterior samples with
Gaussian Mixture Models.

Copyright(C) 2024 by
Trey V. Wenger; tvwenger@gmail.com
This code is licensed under MIT license (see LICENSE for details)
"""

from typing import Iterable

import arviz as az
import numpy as np
import xarray
from scipy.stats import mode
from sklearn.mixture import GaussianMixture


def cluster_posterior(
    trace: az.InferenceData,
    n_clusters: int,
    cluster_features: Iterable[str],
    num_gmm_samples: float = 10_000,
    kl_div_threshold: float = 0.1,
    seed: int = 1234,
) -> list:
    """Identify unique solutions and break the labeling degeneracy. To do so, we
    (1) fit a Gaussian Mixture Model (GMM) to the posterior samples of each chain individually.
    (2) calculate the Kullback–Leibler (KL) divergence (mean log-likelihood ratio) between
    chains. If the KD divergence is smaller than the given threshold, then both chains are
    part of the same solution. Otherwise, then each chain belongs to a different solution.
    The KL divergence is calculated from samples drawn from the fitted GMMs following the
    Monte Carlo procedure of Hershey & Olson (2007)
    (3) solve the labeling degeneracy by identifying the most common order of components
    among chains in each solution.

    :param trace: Posterior samples
    :type trace: az.InferenceData
    :param n_clusters: Number of GMM clusters
    :type n_clusters: int
    :param cluster_features: Parameter names to use for clustering
    :type cluster_features: Iterable[str]
    :param num_gmm_samples: Number of samples to generate from Gaussian Mixture Model (GMM), defaults to 10_000
    :type num_gmm_samples: int, optional
    :param kl_div_threshold: Kullback-Liebler (KL) divergence threshold, defaults to 0.1
    :type kl_div_threshold: float, optional
    :param seed: Random seed, defaults to 1234
    :type seed: int, optional
    :return: Solutions, where each element is a dictionary containing posterior samples and other statistics
    :rtype: list
    """
    # Fit GMMs to posterior samples of each chain, generate samples from GMMs
    gmm_results = {}
    for chain in trace.chain.data:
        features = np.array(
            [trace[param].sel(chain=chain).data.flatten() for param in cluster_features]
        ).T
        gmm = GaussianMixture(
            n_components=n_clusters,
            max_iter=100,
            init_params="random_from_data",
            n_init=10,
            verbose=False,
            random_state=seed,
        )
        gmm.fit(features)
        gmm_results[chain] = {"gmm": gmm, "samples": gmm.sample(num_gmm_samples)[0]}

    # Evaluate pair-wise KL divergence
    for chain1 in trace.chain.data:
        gmm_results[chain1]["kl_div"] = {}
        samples = gmm_results[chain1]["samples"]
        lnlike1 = gmm_results[chain1]["gmm"].score(samples)
        for chain2 in trace.chain.data:
            lnlike2 = gmm_results[chain2]["gmm"].score(samples)
            gmm_results[chain1]["kl_div"][chain2] = np.mean(lnlike1 - lnlike2)

    # Group unique solutions based on KL divergence
    solutions = []
    assigned_chains = []
    for chain1 in trace.chain.data:
        # skip if this chain is already assigned
        if chain1 in assigned_chains:
            continue

        # identify chains within KL divergence threshold
        good_kl_div_chains = [
            ch
            for ch, kl_div in gmm_results[chain1]["kl_div"].items()
            if kl_div < kl_div_threshold and ch not in assigned_chains
        ]

        # skip if fewer than two chains are assigned to this solution
        if len(good_kl_div_chains) < 2:  # pragma: no cover
            continue

        # get label order of each chain based on this chain's GMM
        solution = {"label_orders": {}}
        for chain2 in good_kl_div_chains:
            features = np.array(
                [
                    trace[param].sel(chain=chain2).data.flatten()
                    for param in cluster_features
                ]
            ).T
            labels = (
                gmm_results[chain1]["gmm"].predict(features).reshape(-1, n_clusters)
            )
            label_order = mode(labels, axis=0).mode

            # ensure all labels present
            if len(np.unique(label_order)) == len(label_order):
                solution["label_orders"][chain2] = label_order

        # save solution only if at least two chains have all labels present
        if len(solution["label_orders"]) >= 2:
            solutions.append(solution)

            # these chains are now assigned
            assigned_chains += list(solution["label_orders"].keys())

    # Each solution now has the labeling degeneracy broken, in that
    # each cloud has been assigned to a unique GMM cluster. We must
    # now determine which order of GMM clusters is preferred
    good_solutions = []
    for solution in solutions:
        label_orders = np.array(
            [label_order for label_order in solution["label_orders"].values()]
        )
        unique_label_orders, counts = np.unique(
            label_orders,
            axis=0,
            return_counts=True,
        )
        solution["label_order"] = unique_label_orders[np.argmax(counts)]

        # Determine the order of clouds needed to match the adopted label order.
        solution["cloud_orders"] = {}
        for chain, label_order in solution["label_orders"].items():
            xorder = np.argsort(label_order)
            solution["cloud_orders"][chain] = xorder[
                np.searchsorted(label_order[xorder], solution["label_order"])
            ]

        # Add solutions to the trace
        coords = trace.coords.copy()
        coords["chain"] = list(solution["cloud_orders"].keys())
        dims = {}
        posterior_clustered = {}
        for param, samples in trace.data_vars.items():
            if "cloud" in samples.coords:
                # break labeling degeneracy
                posterior_clustered[param] = xarray.concat(
                    [
                        samples.sel(chain=chain, cloud=cloud_order).assign_coords(
                            cloud=range(n_clusters)
                        )
                        for chain, cloud_order in solution["cloud_orders"].items()
                    ],
                    dim="chain",
                )
            else:
                posterior_clustered[param] = xarray.concat(
                    [
                        samples.sel(chain=chain)
                        for chain in solution["cloud_orders"].keys()
                    ],
                    dim="chain",
                )
            dims[param] = list(samples.dims)
        solution["posterior_clustered"] = posterior_clustered
        solution["coords"] = coords
        solution["dims"] = dims
        good_solutions.append(solution)
    return good_solutions



"""
gauss_noise_model.py
Defines GaussNoiseModel, a Gaussian line profile model.

Copyright(C) 2024 by
Trey V. Wenger; tvwenger@gmail.com
This code is licensed under MIT license (see LICENSE for details)
"""

import pymc as pm

from bayes_spec.models import GaussModel


class GaussNoiseModel(GaussModel):
    """
    Definition of a Gaussian line profile model, with noise as an additional
    free parameter.
    """

    def __init__(self, *args, **kwargs):
        """Initialize a new GaussModel instance

        :param `*args`: Arguments passed to :class:`BaseModel`
        :param `**kwargs`: Keyword arguments passed to :class:`BaseModel`
        """
        # Initialize BaseModel
        super().__init__(*args, **kwargs)

        # Define TeX representation of each parameter
        self.var_name_map.update(
            {
                "rms_observation": r"rms (K)",
            }
        )

    def add_priors(
        self,
        prior_rms: float = 1.0,
        **kwargs,
    ):
        """Add priors to the model.

        :param prior_rms: Prior distribution on spectral rms (K), where
            rms ~ HalfNormal(sigma=prior_rms)
            defaults to 1.0
        :type prior_rms: float, optional
        :param `**kwargs`: Additional keyword arguments passed to :class:`GaussModel.add_priors`
        """
        # Add GaussLine priors
        super().add_priors(**kwargs)

        # check inputs
        if not isinstance(prior_rms, float):
            raise ValueError("prior_rms must be a number")

        # Add additional priors
        with self.model:
            # Spectral rms (K)
            rms_observation_norm = pm.HalfNormal("rms_observation_norm", sigma=1.0)
            _ = pm.Deterministic("rms_observation", rms_observation_norm * prior_rms)

    def add_likelihood(self):
        """Add likelihood to the model. Data key must be "observation"."""
        # Predict emission
        predicted = self.predict()

        with self.model:
            # Evaluate likelihood
            _ = pm.Normal(
                "observation",
                mu=predicted,
                sigma=self.model["rms_observation"],
                observed=self.data["observation"].brightness,
            )



"""
plots.py - Plotting helper utilities.

Copyright(C) 2024 by
Trey V. Wenger; tvwenger@gmail.com
This code is licensed under MIT license (see LICENSE for details)
"""

import warnings
from typing import Optional, Iterable

import arviz as az
import arviz.labels as azl

from matplotlib.axes import Axes
import matplotlib.pyplot as plt
import numpy as np

from bayes_spec import SpecData


def plot_traces(posterior: az.InferenceData, var_names: list[str]) -> Iterable[Axes]:
    """Helper function to generate trace plots of posterior samples

    :param posterior: Posterior samples
    :type posterior: az.InferenceData
    :param var_names: Parameters to plot
    :type var_names: list[str]
    :return: `matplotlib` `Axes`
    :rtype: Axes
    """
    with az.rc_context(rc={"plot.max_subplots": None}):
        axes = az.plot_trace(
            posterior,
            var_names=var_names,
        )
    return axes


def plot_predictive(
    data: dict[str, SpecData],
    predictive: az.InferenceData,
) -> Iterable[Axes]:
    """Helper function to generate posterior predictive check plots.

    :param data: Data sets, where the key defines the name of the dataset.
    :type data: dict[str, SpecData]
    :param predictive: Predictive samples
    :type predictive: az.InferenceData
    :return: `matplotlib` `Axes`
    :rtype: Axes
    """
    fig, axes = plt.subplots(len(data), squeeze=False, layout="constrained")
    num_chains = len(predictive.chain)

    # Loop over datasets
    for idx, (key, dataset) in enumerate(data.items()):
        color = iter(plt.cm.rainbow(np.linspace(0, 1, num_chains)))
        # Loop over chains
        for chain in predictive.chain:
            c = next(color)
            # plot predictives
            outcomes = predictive[key].sel(chain=chain).data
            axes[idx][0].plot(
                dataset.spectral,
                outcomes.T,
                linestyle="-",
                color=c,
                alpha=0.5,
                linewidth=2,
            )

        # plot data
        axes[idx][0].plot(
            dataset.spectral,
            dataset.brightness,
            "k-",
        )
        axes[idx][0].set_xlabel(dataset.xlabel)
        axes[idx][0].set_ylabel(dataset.ylabel)
    return axes


def plot_pair(
    trace: az.InferenceData,
    var_names: list[str],
    labeller: Optional[azl.MapLabeller] = None,
    kind: str = "scatter",
    reference_values: Optional[dict] = None,
    kde_kwargs: Optional[dict] = None,
    scatter_kwargs: Optional[dict] = None,
    hexbin_kwargs: Optional[dict] = None,
    reference_values_kwargs: Optional[dict] = None,
) -> Iterable[Axes]:
    """Helper function to generate sample pair plots.

    :param trace: Samples
    :type trace: az.InferenceData
    :param var_names: Parameter names to plot
    :type var_names: list[str]
    :param labeller: `arviz` labeler, defaults to None
    :type labeller: Optional[azl.MapLabeller], optional
    :param kind: plot kind, one of "scatter", "hexbin", or "kde", defaults to "scatter"
    :type kind: str
    :param reference_values: highlight reference values, defaults to None
    :param reference_values: Optional[dict], optional
    :param kde_kwargs: keyword arguments for arviz.plot_kde(), defaults to None
    :param kde_kwargs: Optional[dict], optional
    :param scatter_kwargs: keyword arguments for plt.scatter(), defaults to None
    :param scatter_kwargs: Optional[dict], optional
    :param hexbin_kwargs: keyword arguments for plt.hexbin(), defaults to None
    :param hexbin_kwargs: Optional[dict], optional
    :param reference_values_kwargs: keyword arguments for plt.scatter(), defaults to None
    :param reference_values_kwargs: Optional[dict], optional
    :return: `matplotlib` `Axes`
    :rtype: Axes
    """
    if kde_kwargs is None:
        kde_kwargs = {
            "hdi_probs": [
                0.3,
                0.6,
                0.9,
            ],  # Plot 30%, 60% and 90% HDI contours
            "contourf_kwargs": {"cmap": "Grays"},
            "contour_kwargs": {"colors": "k"},
        }
    if scatter_kwargs is None:
        scatter_kwargs = {
            "marker": ".",
            "color": "k",
            "alpha": 0.1,
        }
    if reference_values_kwargs is None:
        reference_values_kwargs = {
            "marker": "o",
            "color": "r",
            "markersize": 10,
            "linestyle": "none",
        }

    size = int(2.0 * (len(var_names) + 1))
    textsize = int(np.sqrt(size)) + 8
    with az.rc_context(rc={"plot.max_subplots": None}):
        with warnings.catch_warnings(action="ignore"):
            axes = az.plot_pair(
                trace,
                var_names=var_names,
                combine_dims={"cloud"},
                kind=kind,
                figsize=(size, size),
                labeller=labeller,
                marginals=True,
                marginal_kwargs={"color": "k"},
                textsize=textsize,
                kde_kwargs=kde_kwargs,
                scatter_kwargs=scatter_kwargs,
                hexbin_kwargs=hexbin_kwargs,
                backend_kwargs={"layout": "constrained"},
                reference_values=reference_values,
                reference_values_kwargs=reference_values_kwargs,
            )

    # drop y-label of top left marginal
    axes[0][0].set_ylabel("")

    # drop grid lines
    for ax in axes.flatten():
        ax.grid(False)
    return axes



__all__ = [
    "SpecData",
    "BaseModel",
    "Optimize",
]

from bayes_spec.spec_data import SpecData
from bayes_spec.base_model import BaseModel
from bayes_spec.optimize import Optimize

from . import _version
__version__ = _version.get_versions()['version']



"""
spec_data.py - SpecData structure definition

Copyright(C) 2024 by
Trey V. Wenger; tvwenger@gmail.com
This code is licensed under MIT license (see LICENSE for details)
"""

import numpy as np


class SpecData:
    """
    SpecData defines the data structure and utility functions.
    """

    def __init__(
        self,
        spectral: list[float],
        brightness: list[float],
        noise: float | list[float],
        xlabel: str = "Spectral",
        ylabel: str = "Brightness",
    ):
        """Initialize a new `SpecData` instance.

        :param spectral: Spectaral axis
        :type spectral: list[float]
        :param brightness: Brightness data
        :type brightness: list[float]
        :param noise: Noise data. If a scalar, then the noise is assumed constant across the spectrum.
        :type noise: float | list[float]
        :param xlabel: Label for spectral axis, defaults to "Spectral"
        :type xlabel: str, optional
        :param ylabel: Label for brightness axis, defaults to "Brightness"
        :type ylabel: str, optional
        :raises ValueError: Size mis-match between :param:`spectral` and :param:`brightness`
        :raises ValueError: :param:`noise` is not a scalar and there is a size mis-match between :param:`brightness` and :param:`noise`
        """
        if len(spectral) != len(brightness):
            raise ValueError("size mismatch between brightness and spectral")
        self.spectral = spectral
        self.brightness = brightness
        self.xlabel = xlabel
        self.ylabel = ylabel

        if len(np.atleast_1d(noise)) == 1:
            self.noise = np.ones_like(brightness) * noise
        elif len(noise) != len(self.brightness):
            raise ValueError("size mismatch between brightness and noise")
        else:
            self.noise = noise

        # normalize spectral axis to unit domain [-1, 1]
        self._spectral_offset = np.mean(self.spectral)
        self._spectral_scale = np.ptp(self.spectral) / 2.0
        self.spectral_norm = self.normalize_spectral(self.spectral)

        # normalize brightness data using standard normalization
        self._brightness_offset = np.median(self.brightness)
        self._brightness_scale = np.median(self.noise)
        self.brightness_norm = self.normalize_brightness(brightness)

    def _normalize(self, x: float, offset: float, scale: float) -> float:
        """Normalize some data

        :param x: data to normalize
        :type x: float
        :param offset: Normalization offset
        :type offset: float
        :param scale: Normalization scale
        :type scale: float
        :return: Normalized data
        :rtype: float
        """
        return (x - offset) / scale

    def _unnormalize(self, norm_x: float, offset: float, scale: float) -> float:
        """Un-normalize some data

        :param norm_x: Normalized data
        :type norm_x: float
        :param offset: Normalization offset
        :type offset: float
        :param scale: Normalization scale
        :type scale: float
        :return: Un-normalized data
        :rtype: float
        """
        return norm_x * scale + offset

    def normalize_spectral(self, x: float) -> float:
        """Normalize spectral data

        :param x: Spectral data to normalize
        :type x: float
        :return: Normalized spectral data
        :rtype: float
        """
        return self._normalize(x, self._spectral_offset, self._spectral_scale)

    def unnormalize_spectral(self, norm_x: float) -> float:
        """Un-normalize spectral data

        :param norm_x: Normalized spectral data
        :type norm_x: float
        :return: Un-normalized spectral data
        :rtype: float
        """
        return self._unnormalize(norm_x, self._spectral_offset, self._spectral_scale)

    def normalize_brightness(self, x: float) -> float:
        """Normalize brightness data

        :param x: Brightness data to normalize
        :type x: float
        :return: Normalized brightness data
        :rtype: float
        """
        return self._normalize(x, self._brightness_offset, self._brightness_scale)

    def unnormalize_brightness(self, norm_x: float) -> float:
        """Un-normalize brighrtness data

        :param norm_x: Normalized brightness data
        :type norm_x: float
        :return: Un-normalized brightness data
        :rtype: float
        """
        return self._unnormalize(norm_x, self._brightness_offset, self._brightness_scale)



"""
base_model.py - BaseModel definition

Copyright(C) 2024 by
Trey V. Wenger; tvwenger@gmail.com
This code is licensed under MIT license (see LICENSE for details)
"""

import warnings
from abc import ABC, abstractmethod
from typing import Optional, Iterable

import pymc as pm
from pymc.variational.callbacks import CheckParametersConvergence
from pymc.model.transform.optimization import freeze_dims_and_data

import pytensor.tensor as pt

import arviz as az
import arviz.labels as azl
import graphviz

import numpy as np
from numpy.polynomial import Polynomial

from scipy.stats import norm

from bayes_spec import SpecData
from bayes_spec.cluster_posterior import cluster_posterior
from bayes_spec.nuts import init_nuts


class BaseModel(ABC):
    """
    BaseModel defines functions and attributes common to all model definitions.
    """

    def __init__(
        self,
        data: dict[str, SpecData],
        n_clouds: int,
        baseline_degree: int = 0,
        seed: int = 1234,
        verbose: bool = False,
    ):
        """Initialize a new BaseModel

        :param data: Spectral data sets, where the "key" defines the name of the dataset
        :type data: dict[str, SpecData]
        :param n_clouds: Number of cloud components
        :type n_clouds: int
        :param baseline_degree: Polynomial baseline degree, defaults to 0
        :type baseline_degree: int, optional
        :param seed: Random seed, defaults to 1234
        :type seed: int, optional
        :param verbose: Print verbose output, defaults to False
        :type verbose: bool, optional
        """
        self.n_clouds = n_clouds
        self.baseline_degree = baseline_degree
        self.seed = seed
        self.verbose = verbose
        self.data = data

        # Initialize the model
        coords = {
            "baseline_coeff": range(self.baseline_degree + 1),
            "cloud": range(self.n_clouds),
        }
        self._n_data = 0
        for _, dataset in self.data.items():
            self._n_data += len(dataset.spectral)
        self.model = pm.Model(coords=coords)

        # Parameters used for posterior clustering
        self._cluster_features = []

        # Arviz labeller map
        self.var_name_map = {
            f"baseline_{key}_norm": r"$\beta_{\rm " + key + r"}$"
            for key in self.data.keys()
        }

        # set results and convergence checks
        self.reset_results()

    @abstractmethod
    def add_priors(self, *args, **kwargs):  # pragma: no cover
        """Must be defined in inhereted class."""
        pass

    @abstractmethod
    def add_likelihood(self, *args, **kwargs):  # pragma: no cover
        """Must be defined in inhereted class."""
        pass

    def _get_param_names(
        self, dim: Optional[str] = None, deterministics: bool = False
    ) -> Iterable[str]:
        """Get a subset of the model parameter names.

        :param dims: Select only parameters with this dimension name. If None, select dimensionless
            parameters, defaults to None
        :type dims: Optional[str], optional
        :param deterministics: If True, select deterministic parameters, otherwise free
            parameters, defaults to False
        :type deterministics: bool, optional
        :return: Parameter names matching selection
        :rtype: Iterable[str]
        """
        all_params = (
            self.model.deterministics if deterministics else self.model.free_RVs
        )
        if dim is None:
            return [
                param.name
                for param in all_params
                if param.name not in self.model.named_vars_to_dims
            ]
        return [
            param.name
            for param in all_params
            if dim in self.model.named_vars_to_dims.get(param.name, [])
        ]

    @property
    def baseline_freeRVs(self) -> Iterable[str]:
        """Get the free baseline parameter names.

        :return: Free baseline parameter names
        :rtype: Iterable[str]
        """
        return self._get_param_names("baseline_coeff", deterministics=False)

    @property
    def baseline_deterministics(self) -> Iterable[str]:
        """Get the deterministic baseline parameter names.

        :return: Deterministic baseline parameter names
        :rtype: Iterable[str]
        """
        return self._get_param_names("baseline_coeff", deterministics=True)

    @property
    def cloud_freeRVs(self) -> Iterable[str]:
        """Get the free cloud parameter names.

        :return: Free cloud parameter names
        :rtype: Iterable[str]
        """
        return self._get_param_names("cloud", deterministics=False)

    @property
    def cloud_deterministics(self) -> Iterable[str]:
        """Get the deterministic cloud parameter names.

        :return: Deterministic cloud parameter names
        :rtype: Iterable[str]
        """
        return self._get_param_names("cloud", deterministics=True)

    @property
    def hyper_freeRVs(self) -> Iterable[str]:
        """Get the free hyper parameter names.

        :return: Free hyper parameter names
        :rtype: Iterable[str]
        """
        return self._get_param_names(deterministics=False)

    @property
    def hyper_deterministics(self) -> Iterable[str]:
        """Get the deterministic hyper parameter names.

        :return: Deterministic hyper parameter names
        :rtype: Iterable[str]
        """
        return self._get_param_names(deterministics=True)

    @property
    def _n_params(self) -> int:
        """Determine the number of free model parameters.

        :return: Number of free model parameters
        :rtype: int
        """
        return np.sum([param.size.eval().prod() for param in self.model.free_RVs])

    @property
    def _get_unique_solution(self) -> int:
        """Return the unique solution index (0) if there is a unique solution, otherwise raise an exception.

        :raises ValueError: No unique solution
        :return: Unique solution index (`0`)
        :rtype: int
        """
        if not self.unique_solution:  # pragma: no cover
            raise ValueError("There is not a unique solution. Must supply solution.")
        return 0

    @property
    def unique_solution(self) -> bool:
        """Check if posterior samples suggest a unique solution.

        :raises ValueError: No solutions
        :return: True if there is a unique solution, False otherwise
        :rtype: bool
        """
        if self.solutions is None or len(self.solutions) == 0:
            raise ValueError("No solutions. Try solve()")
        return len(self.solutions) == 1

    @property
    def labeller(self) -> azl.MapLabeller:
        """Get the arviz labeller.

        :return: Arviz labeller
        :rtype: azl.MapLabeller
        """
        return azl.MapLabeller(var_name_map=self.var_name_map)

    def _validate(self) -> bool:
        """Validate the model by checking the log probability at the initial point.

        :raises ValueError: Model does not contain likelihood
        :raises ValueError: Model likelihood fails to evaluate at the initial point
        :return: True
        :rtype: bool
        """
        # check that likelihood has been added
        if len(self.model.observed_RVs) == 0:
            raise ValueError("No observed variables found! Did you add_likelihood()?")

        # check that model can be evaluated
        if not np.isfinite(self.model.logp().eval(self.model.initial_point())):
            raise ValueError(
                "Model initial point is not finite! Mis-specified model or bad priors?"
            )

        return True

    def reset_results(self):
        """Reset results and convergence checks."""
        self.approx: pm.Approximation = None
        self.trace: az.InferenceData = None
        self.solutions: list = None
        self._chains_converged: bool = None

    def graph(self) -> graphviz.sources.Source:
        """Generate visualization of the model graph. The output can be displayed in-line in a Jupyter notebook,
        or rendered with `graph().render('filename')`.

        :return: Graph visualization
        :rtype: graphviz.sources.Source
        """
        gviz = pm.model_to_graphviz(self.model)
        gviz.graph_attr["rankdir"] = "TB"
        gviz.graph_attr["splines"] = "ortho"
        gviz.graph_attr["newrank"] = "false"
        source = gviz.unflatten(stagger=3)
        return source

    def null_bic(self) -> float:
        """Evaluate the Bayesian Information Criterion for the null hypothesis (baseline only, no clouds)

        :return: Null hypothesis BIC
        :rtype: float
        """
        lnlike = 0.0
        for _, dataset in self.data.items():
            # fit polynomial baseline to un-normalized spectral data
            baseline = Polynomial.fit(
                dataset.spectral, dataset.brightness, self.baseline_degree
            )(dataset.spectral)

            # evaluate likelihood
            lnlike += norm.logpdf(
                dataset.brightness - baseline, scale=dataset.noise
            ).sum()

        n_params = len(self.data) * (self.baseline_degree + 1)
        return n_params * np.log(self._n_data) - 2.0 * lnlike

    def mean_lnlike(
        self, chain: Optional[int] = None, solution: Optional[int] = None
    ) -> float:
        """Evaluate mean log-likelihood over posterior samples.

        :param chain: Evaluate mean log-likelihood for this chain using un-clustered posterior samples.
            If `None` evaluate across all chains using clustered posterior samples, defaults to None
        :type chain: Optional[int], optional
        :param solution: Evaluate mean log-likelihood for this solution. If `None` use the unique solution if any.
            If :param:chain is not None, this parameter has no effect, defaults to None
        :type solution: Optional[int], optional
        :return: Mean log-likelihood over posterior samples
        :rtype: float
        """
        if chain is None and solution is None:
            solution = self._get_unique_solution

        # likelihood samples
        if chain is None:
            chain = self.trace[f"solution_{solution}"].chain
        lnlike = self.trace["log_likelihood"].sel(chain=chain)

        spec_dims = [f"{key}_dim_0" for key in self.data.keys()]
        spec_sum_lnlike = lnlike.sum(dim=spec_dims)
        sum_lnlike = np.sum([spec_sum_lnlike[key] for key in self.data.keys()], axis=0)
        return np.mean(sum_lnlike)

    def bic(self, chain: Optional[int] = None, solution: Optional[int] = None) -> float:
        """Calculate the Bayesian information criterion at the mean point estimate.

        :param chain: Evaluate BIC for this chain using un-clustered posterior samples. If `None` evaluate across all
            chains using clustered posterior samples, defaults to None
        :type chain: Optional[int], optional
        :param solution: Evaluate BIC for this solution. If `None` use the unique solution if any. If :param:chain
            is not None, this parameter has no effect, defaults to None
        :type solution: Optional[int], optional
        :return: Bayesian information criterion
        :rtype: float
        """
        try:
            lnlike = self.mean_lnlike(chain=chain, solution=solution)
            return self._n_params * np.log(self._n_data) - 2.0 * lnlike
        except ValueError as e:  # pragma: no cover
            print(e)
            return np.inf

    def add_baseline_priors(
        self, prior_baseline_coeffs: Optional[dict[str, list[float]]] = None
    ):
        """Add baseline priors to the model. The polynomial baseline is evaluated on the normalized data like:
        `baseline_norm = sum_i(coeff[i]/(i+1)**i * spectral_norm**i)`

        :param prior_baseline_coeffs: Width of normal prior distribution on the normalized baseline polynomial
            coefficients. Keys are dataset names and values are lists of length `baseline_degree+1`. If None,
            use `[1.0]*(baseline_degree+1)` for each dataset, defaults to None
        :type prior_baseline_coeffs: Optional[dict[str, list[float]]], optional
        """
        if prior_baseline_coeffs is None:
            prior_baseline_coeffs = {
                key: [1.0] * (self.baseline_degree + 1) for key in self.data.keys()
            }
        for key, coeffs in prior_baseline_coeffs.items():
            if len(coeffs) != self.baseline_degree + 1:
                raise ValueError(
                    f"{key} baseline coefficient prior must have length "
                    + f"`baseline_degree+1` = {self.baseline_degree + 1}"
                )

        with self.model:
            for key in self.data.keys():
                # add the normalized prior
                _ = pm.Normal(
                    f"baseline_{key}_norm",
                    mu=0.0,
                    sigma=prior_baseline_coeffs[key],
                    dims="baseline_coeff",
                )

    def predict_baseline(
        self, baseline_params: Optional[dict[str, list[float]]] = None
    ) -> dict[str, list[float]]:
        """Predict the un-normalized baseline model.

        :param baseline_params: Dictionary of baseline parameters with which to evaluate the baseline model.
            Keys are the same as in the model: "baseline_{key}_norm", where {key} are the supplied datasets. The
            values are lists of length `baseline_degree+1`. If None, evaluate the baseline model using the current
            model state, defaults to None
        :type baseline_params: Optional[dict[str, list[float]]], optional
        :return: Un-normalized baseline models for each dataset. Keys are dataset names and values are
            the un-normalized baseline models.
        :rtype: dict[str, list[float]]
        """
        if baseline_params is None:
            baseline_params = self.model
        baseline_model = {}
        for key, dataset in self.data.items():
            # evaluate the baseline
            baseline_norm = pt.sum(
                [
                    baseline_params[f"baseline_{key}_norm"][i]
                    / (i + 1.0) ** i
                    * dataset.spectral_norm**i
                    for i in range(self.baseline_degree + 1)
                ],
                axis=0,
            )
            baseline_model[key] = dataset.unnormalize_brightness(baseline_norm)
        return baseline_model

    def sample_prior_predictive(self, samples: int = 50) -> az.InferenceData:
        """Generate prior predictive samples

        :param samples: Number of prior predictive samples to draw, defaults to 50
        :type samples: int, optional
        :return: Prior predictive samples
        :rtype: az.InferenceData
        """
        # validate
        assert self._validate()

        with self.model:
            trace = pm.sample_prior_predictive(samples=samples, random_seed=self.seed)

        return trace

    def sample_posterior_predictive(
        self,
        solution: Optional[int] = None,
        thin: int = 100,
    ) -> az.InferenceData:
        """Generate posterior predictive samples

        :param solution: Draw posterior predictive samples from this solution index. If None, draw samples from
            the un-clustered posterior samples, defaults to None
        :type solution: Optional[int], optional
        :param thin: Thin posterior samples by keeping one in :param:thin, defaults to 100
        :type thin: int, optional
        :raises ValueError: No posterior samples
        :return: Posterior predictive samples
        :rtype: az.InferenceData
        """
        # validate
        assert self._validate()

        if self.trace is None:
            raise ValueError("Model has no posterior samples. Try fit() or sample().")

        with self.model:
            if solution is None:
                posterior = self.trace.posterior.sel(draw=slice(None, None, thin))
            else:
                posterior = self.trace[f"solution_{solution}"].sel(
                    draw=slice(None, None, thin)
                )
            trace = pm.sample_posterior_predictive(
                posterior,
                extend_inferencedata=True,
                random_seed=self.seed,
            )

        return trace

    def fit(
        self,
        n: int = 100_000,
        draws: int = 1_000,
        rel_tolerance: float = 0.001,
        abs_tolerance: float = 0.001,
        learning_rate: float = 1e-3,
        **kwargs,
    ):
        """Approximate posterior distribution using Variational Inference (VI).

        :param n: Number of VI iterations, defaults to 100_000
        :type n: int, optional
        :param draws: Number of posterior samples to draw, defaults to 1_000
        :type draws: int, optional
        :param rel_tolerance: Relative parameter tolerance for VI convergence, defaults to 0.001
        :type rel_tolerance: float, optional
        :param abs_tolerance: Absolute parameter tolerance for VI convergence, defaults to 0.001
        :type abs_tolerance: float, optional
        :param learning_rate: VI learning rate, defaults to 1e-3
        :type learning_rate: float, optional
        :param `**kwargs`: Additional arguments passed to :func:`pymc.fit`
        """
        # validate
        assert self._validate()

        # reset convergence checks
        self.reset_results()

        with self.model:
            callbacks = [
                CheckParametersConvergence(tolerance=rel_tolerance, diff="relative"),
                CheckParametersConvergence(tolerance=abs_tolerance, diff="absolute"),
            ]
            self.approx = pm.fit(
                n=n,
                random_seed=self.seed,
                progressbar=self.verbose,
                callbacks=callbacks,
                obj_optimizer=pm.adagrad_window(learning_rate=learning_rate),
                **kwargs,
            )
            self.trace = self.approx.sample(draws)

        with self.model:
            pm.compute_log_likelihood(self.trace)

    def sample(
        self,
        init: str = "advi+adapt_diag",
        n_init: int = 100_000,
        chains: int = 4,
        init_kwargs: Optional[dict] = None,
        nuts_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        """Sample posterior distribution using MCMC.

        :param init: Initialization strategy, defaults to "advi+adapt_diag"
        :type init: str, optional
        :param n_init: Number of initialization iterations, defaults to 100_000
        :type n_init: int, optional
        :param chains: Number of independent Markov chains, defaults to 4
        :type chains: int, optional
        :param init_kwargs: Keyword arguments passed to :func:`init_nuts`, defaults to None
        :type init_kwargs: Optional[dict], optional
        :param nuts_kwargs: Keyword arguments passed to :func:`pymc.NUTS`, defaults to None
        :type nuts_kwargs: Optional[dict], optional
        :param `**kwargs`: Additional arguments passed to :func:`pymc.sample`
        """
        # validate
        assert self._validate()

        # reset convergence checks
        self.reset_results()

        # catch non-standard initialization for non-pymc samplers
        if (
            "nuts_sampler" in kwargs.keys()
            and kwargs["nuts_sampler"] != "pymc"
            and init != "auto"
        ):
            init = "auto"
            warnings.warn("setting init='auto' for non-pymc sampler")

        if init == "auto":
            init = "jitter+adapt_diag"

        if init_kwargs is None:
            init_kwargs = {}
        if nuts_kwargs is None:
            nuts_kwargs = {}

        # freeze model for JAX samplers
        frozen_model = self.model
        if "nuts_sampler" in kwargs.keys() and kwargs["nuts_sampler"] in [
            "numpyro",
            "blackjax",
        ]:
            frozen_model = freeze_dims_and_data(self.model)

        # attempt custom initialization
        initial_points, step = init_nuts(
            self.model,
            init=init,
            n_init=n_init,
            chains=chains,
            nuts_kwargs=nuts_kwargs,
            seed=self.seed,
            verbose=self.verbose,
            **init_kwargs,
        )

        # if we're using custom initialization, then drop nuts
        # arguments from pm.sample
        if initial_points is not None:
            nuts_kwargs = {}

        with frozen_model:
            self.trace = pm.sample(
                init=init,
                initvals=initial_points,
                step=step,
                chains=chains,
                progressbar=self.verbose,
                discard_tuned_samples=False,
                compute_convergence_checks=False,
                random_seed=self.seed,
                **nuts_kwargs,
                **kwargs,
            )

        # add log-likelihood to trace
        if self.verbose:
            print("Adding log-likelihood to trace")
        with frozen_model:
            pm.compute_log_likelihood(self.trace)

        # diagnostics
        if self.verbose:
            # divergences
            num_divergences = self.trace.sample_stats.diverging.data.sum()
            if num_divergences > 0:  # pragma: no cover
                print(f"There were {num_divergences} divergences in converged chains.")

    def sample_smc(
        self,
        **kwargs,
    ):
        """Sample posterior distribution using Sequential Monte Carlo.

        :param `**kwargs`: Additional arguments passed to :func:`pymc.sample_smc`
        """

        # validate
        assert self._validate()

        # reset convergence checks
        self.reset_results()

        with self.model:
            self.trace = pm.sample_smc(
                progressbar=self.verbose,
                compute_convergence_checks=False,
                **kwargs,
            )

        # add log-likelihood to trace
        if self.verbose:
            print("Adding log-likelihood to trace")
        with self.model:
            pm.compute_log_likelihood(self.trace, progressbar=self.verbose)

    def solve(self, num_gmm_samples: float = 10_000, kl_div_threshold: float = 0.1):
        """Identify unique solutions and break the labeling degeneracy.
        Adds new groups to the `trace` called `solution_{idx}` with the label-corrected posterior
        samples of each unique solution.

        :param num_gmm_samples: Number of samples to generate from Gaussian Mixture Model (GMM), defaults to 10_000
        :type num_gmm_samples: int, optional
        :param kl_div_threshold: Kullback-Liebler (KL) divergence threshold, defaults to 0.1
        :type kl_div_threshold: float, optional
        """
        # Drop solutions if they already exist in trace
        for group in list(self.trace.groups()):
            if "solution" in group:  # pragma: no cover
                del self.trace[group]

        self.solutions = []
        solutions = cluster_posterior(
            self.trace.posterior,
            self.n_clouds,
            self._cluster_features,
            num_gmm_samples=num_gmm_samples,
            kl_div_threshold=kl_div_threshold,
            seed=self.seed,
        )
        if len(solutions) < 1 and self.verbose:  # pragma: no cover
            print("No solution found!")

        # convergence checks
        unique_solution = len(solutions) == 1
        if self.verbose:
            if unique_solution:
                print("GMM converged to unique solution")
            elif len(solutions) > 1:  # pragma: no cover
                print(f"GMM found {len(solutions)} unique solutions")
                for solution_idx, solution in enumerate(solutions):
                    print(
                        f"Solution {solution_idx}: chains {list(solution['label_orders'].keys())}"
                    )

        assigned_chains = []
        for solution in solutions:
            assigned_chains += list(solution["label_orders"].keys())
        if self.verbose and len(assigned_chains) < len(
            self.trace.posterior.chain
        ):  # pragma: no cover
            print(
                f"{len(assigned_chains)} of {len(self.trace.posterior.chain)} chains appear converged."
            )

        for solution_idx, solution in enumerate(solutions):
            # report labeling degeneracy
            label_orders = np.array(
                [label_order for label_order in solution["label_orders"].values()]
            )
            if self.verbose and not np.all(
                label_orders == label_orders[0]
            ):  # pragma: no cover
                print(f"Label order mismatch in solution {solution_idx}")
                for chain, label_order in solution["label_orders"].items():
                    print(f"Chain {chain} order: {label_order}")
                print(f"Adopting (first) most common order: {solution['label_order']}")

            # Add solution to the trace
            with warnings.catch_warnings(action="ignore"):
                self.trace.add_groups(
                    **{
                        f"solution_{solution_idx}": solution["posterior_clustered"],
                        "coords": solution["coords"],
                        "dims": solution["dims"],
                    }
                )
                self.solutions.append(solution_idx)



"""
test_optimize.py - Unit tests for Optimize functionality

Copyright(C) 2024 by
Trey V. Wenger; tvwenger@gmail.com
This code is licensed under MIT license (see LICENSE for details)
"""

import numpy as np

from bayes_spec import SpecData, Optimize
from bayes_spec.models import GaussModel

_RNG = np.random.RandomState(seed=1234)

noise = 1.0
spectral = np.linspace(-100.0, 100.0, 1000)
dummy_brightness = noise * _RNG.randn(1000)
dummy_data = {"observation": SpecData(spectral, dummy_brightness, noise)}
params = {
    "line_area": [1000.0],
    "fwhm": [25.0],
    "velocity": [10.0],
    "baseline_observation_norm": [0.0],
}
_MODEL = GaussModel(dummy_data, 1, baseline_degree=0, seed=1234, verbose=True)
_MODEL.add_priors()
_MODEL.add_likelihood()
brightness = _MODEL.model["observation"].eval(params)
_DATA = {"observation": SpecData(spectral, brightness, noise)}


def test_fit_all():
    # Test Optimize.fit_all
    opt = Optimize(GaussModel, _DATA, max_n_clouds=1, verbose=True)
    opt.add_priors()
    opt.add_likelihood()
    start_spread = {"velocity_norm": [-3.0, 3.0]}
    opt.fit_all(start_spread=start_spread)


def test_sample_all():
    # Test Optimize.sample_all
    opt = Optimize(GaussModel, _DATA, max_n_clouds=1, verbose=True)
    opt.add_priors()
    opt.add_likelihood()
    start_spread = {"velocity_norm": [-3.0, 3.0]}
    opt.sample_all(start_spread=start_spread)


def test_sample_smc_all():
    # Test Optimize.sample_smc_all
    opt = Optimize(GaussModel, _DATA, max_n_clouds=1, verbose=True)
    opt.add_priors()
    opt.add_likelihood()
    opt.sample_smc_all()


def test_optimize_vi_mcmc():
    # Test Optimize with VI + MCMC
    opt = Optimize(GaussModel, _DATA, max_n_clouds=5, verbose=True)
    opt.add_priors()
    opt.add_likelihood()
    start_spread = {"velocity_norm": [-3.0, 3.0]}
    opt.optimize(
        start_spread=start_spread,
        approx=True,
        smc=False,
    )
    assert opt.best_model.n_clouds == 1
    assert opt.null_bic == opt.models[1].null_bic()
    assert len(opt.bics) == 6


def test_optimize_mcmc():
    # MCMC only
    opt = Optimize(GaussModel, _DATA, max_n_clouds=1, verbose=True)
    opt.add_priors()
    opt.add_likelihood()
    start_spread = {"velocity_norm": [-3.0, 3.0]}
    opt.optimize(
        start_spread=start_spread,
        approx=False,
        smc=False,
    )


def test_optimize_vi_smc():
    # VI + SMC
    opt = Optimize(GaussModel, _DATA, max_n_clouds=1, verbose=True)
    opt.add_priors()
    opt.add_likelihood()
    sample_kwargs = {
        "chains": 4,
        "draws": 500,
    }
    opt.optimize(
        sample_kwargs=sample_kwargs,
        approx=True,
        smc=True,
    )


def test_optimize_smc():
    # SMC only
    opt = Optimize(GaussModel, _DATA, max_n_clouds=1, verbose=True)
    opt.add_priors()
    opt.add_likelihood()
    sample_kwargs = {
        "chains": 4,
        "draws": 500,
    }
    opt.optimize(
        sample_kwargs=sample_kwargs,
        approx=False,
        smc=True,
    )



"""
optimize.py - Fit spectra with MCMC and determine optimal number of
spectral components.

Copyright(C) 2024 by
Trey V. Wenger; tvwenger@gmail.com
This code is licensed under MIT license (see LICENSE for details)
"""

from typing import Type, Optional, Iterable

import numpy as np

from bayes_spec.base_model import BaseModel
from bayes_spec import SpecData


class Optimize:
    """
    Optimize class definition
    """

    def __init__(
        self,
        model_type: Type[BaseModel],
        data: dict[str, SpecData],
        max_n_clouds: int = 5,
        baseline_degree: int = 0,
        seed: int = 1234,
        verbose: bool = False,
        **kwargs,
    ):
        """Initialize a new Optimize instance.

        :param model_type: Model to optimize
        :type model_type: Type[BaseModel]
        :param data: Spectral data sets, where the "key" defines the name of the dataset
        :type data: dict[str, SpecData]
        :param max_n_clouds: Maximum number of clouds to fit, defaults to 5
        :type max_n_clouds: int, optional
        :param baseline_degree: Polynomial baseline degree, defaults to 0
        :type baseline_degree: int, optional
        :param seed: Random seed, defaults to 1234
        :type seed: int, optional
        :param verbose: Verbose output, defaults to False
        :type verbose: bool, optional
        :param `**kwargs`: Additional keyword arguments passed to :param:`model` initialization
        """
        self.model_type = model_type
        self.max_n_clouds = max_n_clouds
        self.verbose = verbose
        self.n_clouds = [i for i in range(1, self.max_n_clouds + 1)]
        self.seed = seed
        self.data = data

        # Initialize models
        self.models: dict[int, Type[BaseModel]] = {}
        for n_cloud in self.n_clouds:
            self.models[n_cloud] = model_type(
                self.data,
                n_cloud,
                baseline_degree,
                seed=seed,
                verbose=self.verbose,
                **kwargs,
            )
        self.best_model = None

    @property
    def null_bic(self) -> float:
        """Evaluate the Bayesian Information Criterion for the null hypothesis (baseline only, no clouds)

        :return: Null hypothesis BIC
        :rtype: float
        """
        return self.models[1].null_bic()

    @property
    def bics(self) -> dict[int, float]:
        """Return the Bayesian Information Criteria for the best solution of each model.

        :return: BIC for each model, indexed by the number of clouds
        :rtype: dict[int, float]
        """
        model_bics = {0: self.null_bic}
        for n_cloud, model in self.models.items():
            best_solution_bic = np.inf
            # Check that model has been sampled
            if model.trace is not None:
                # VI does not have solution, only single "chain"
                if len(model.trace.posterior.chain) == 1:
                    best_solution_bic = model.bic(chain=[0])
                else:
                    for solution in model.solutions:
                        solution_bic = model.bic(solution=solution)
                        if solution_bic < best_solution_bic:
                            best_solution_bic = solution_bic
            model_bics[n_cloud] = best_solution_bic
        return model_bics

    def _check_stop(self, n_cloud: int, bic_threshold: float = 10.0):
        """Check if any of the stopping criteria are met. Stopping criteria are:
        1. Model did not converge
        2. BIC did not improve by more than bic_threshold over previous model

        :param `n_cloud`: model to check
        :type n_cloud: int
        :param bic_threshold: The `best_model` is the first with BIC within `min(BIC)+bic_threshold`, defaults to 10.0
        :type bic_threshold: float, optional

        :return: True if any stopping criteria are met
        :rtype: bool
        """
        # Check there's a trace
        if self.models[n_cloud].trace is None:  # pragma: no cover
            return True

        # Check if there are no solutions
        if len(self.models[n_cloud].trace.posterior.chain) > 1:
            if (
                self.models[n_cloud].solutions is None
                or len(self.models[n_cloud].solutions) == 0
            ):  # pragma: no cover
                return True

        # Get last non-inf BIC
        bics = self.bics
        last_bic = bics[0]
        for n in range(1, n_cloud):
            model_bic = bics[n]
            if not np.isinf(model_bic):
                last_bic = model_bic

        # Check BIC
        this_bic = bics[n_cloud]
        if this_bic + bic_threshold >= last_bic:
            return True

        # No stopping criteria are met
        return False

    def add_priors(self, *args, **kwargs):
        """Add priors to the models

        :param `*args`: Arguments passed to :func:`model.add_priors`
        :param `**kwargs`: Keyword arguments passed to :func:`model.add_priors`
        """
        for n_cloud in self.n_clouds:
            self.models[n_cloud].add_priors(*args, **kwargs)

    def add_likelihood(self, *args, **kwargs):
        """Add likelihood to the models

        :param `*args`: Arguments passed to :func:`model.add_likelihood`
        :param `**kwargs`: Keyword arguments passed to :func:`model.add_likelihood`
        """
        for n_cloud in self.n_clouds:
            self.models[n_cloud].add_likelihood(*args, **kwargs)

    def fit_all(
        self, start_spread: Optional[dict[str, Iterable[float]]] = None, **kwargs
    ):
        """Fit all models using variational inference.

        :param `start_spread`: Keys are parameter names and values are range, defaults to None
        :type start_spread: Optional[dict[str, Iterable[float]]], optional
        :param `**kwargs`: Keyword arguments passed to :func:`model.fit`
        """
        if self.verbose:
            print(f"Null hypothesis BIC = {self.models[1].null_bic():.3e}")
        if start_spread is not None and "start" not in kwargs:
            kwargs["start"] = {}

        for n_cloud in self.n_clouds:
            if self.verbose:
                print(f"Approximating n_cloud = {n_cloud} posterior...")

            if start_spread is not None:
                for key, value in start_spread.items():
                    kwargs["start"][key] = np.linspace(value[0], value[1], n_cloud)

            self.models[n_cloud].fit(**kwargs)
            if self.verbose:
                bic = self.models[n_cloud].bic(chain=[0])
                print(f"n_cloud = {n_cloud} BIC = {bic:.3e}")
                print()

    def sample_all(
        self,
        kl_div_threshold: float = 0.1,
        start_spread: Optional[dict[str, Iterable[float]]] = None,
        **kwargs,
    ):
        """Sample posterior distribution of all models using MCMC.

        :param `kl_div_threshold`: GMM convergence threshold
        :type kl_div_threshold: float, optional
        :param `start_spread`: Keys are parameter names and values are range, defaults to None
        :type start_spread: Optional[dict[str, Iterable[float]]], optional
        :param `**kwargs`: Keyword arguments passed to :func:`model.sample`
        """
        if self.verbose:
            print(f"Null hypothesis BIC = {self.models[1].null_bic():.3e}")
        if start_spread is not None and "init_kwargs" not in kwargs:
            kwargs["init_kwargs"] = {}
        if start_spread is not None and "start" not in kwargs["init_kwargs"]:
            kwargs["init_kwargs"]["start"] = {}

        for n_cloud in self.n_clouds:
            if self.verbose:
                print(f"Sampling n_cloud = {n_cloud} posterior...")

            if start_spread is not None:
                for key, value in start_spread.items():
                    kwargs["init_kwargs"]["start"][key] = np.linspace(
                        value[0], value[1], n_cloud
                    )

            self.models[n_cloud].sample(**kwargs)
            self.models[n_cloud].solve(kl_div_threshold=kl_div_threshold)
            if self.verbose:
                for solution in self.models[n_cloud].solutions:
                    print(
                        f"n_cloud = {n_cloud} "
                        + f"solution = {solution} "
                        + f"BIC = {self.models[n_cloud].bic(solution=solution):.3e}"
                    )
                print()

    def sample_smc_all(self, kl_div_threshold: float = 0.1, **kwargs):
        """Sample posterior distribution of all models using sequential Monte Carlo.

        :param `kl_div_threshold`: GMM convergence threshold
        :type kl_div_threshold: float, optional
        :param `**kwargs`: Keyword arguments passed to :func:`model.sample_smc`
        """
        if self.verbose:
            print(f"Null hypothesis BIC = {self.models[1].null_bic():.3e}")

        for n_cloud in self.n_clouds:
            if self.verbose:
                print(f"Sampling n_cloud = {n_cloud} posterior...")

            self.models[n_cloud].sample_smc(**kwargs)
            self.models[n_cloud].solve(kl_div_threshold=kl_div_threshold)
            if self.verbose:
                for solution in self.models[n_cloud].solutions:
                    print(
                        f"n_cloud = {n_cloud} "
                        + f"solution = {solution} "
                        + f"BIC = {self.models[n_cloud].bic(solution=solution):.3e}"
                    )
                print()

    def optimize(
        self,
        bic_threshold: float = 10.0,
        kl_div_threshold: float = 0.1,
        fit_kwargs: Optional[dict] = None,
        sample_kwargs: Optional[dict] = None,
        start_spread: Optional[dict[str, Iterable[float]]] = None,
        smc: bool = False,
        approx: bool = True,
    ):
        """Determine optimal number of clouds by minimizing the Bayesian Information Criterion
        using MCMC, Sequntial Monte Carlo, or Variational Inference. Models are sampled in sequential
        order starting with n_clouds = 1 until the stopping criteria are met twice in succession.
        Then, if approx=True, sample the best model using MCMC or SMC and solve the labeling degeneracy.
        Stopping criteria are:
        1. Model did not converge
        2. Model has multiple solutions (excludeing VI results, which only have one chain)
        3. BIC did not improve by more than bic_threshold over previous model

        :param bic_threshold: The `best_model` is the first with BIC within `min(BIC)+bic_threshold`, defaults to 10.0
        :type bic_threshold: float, optional
        :param `kl_div_threshold`: GMM convergence threshold
        :type kl_div_threshold: float, optional
        :param fit_kwargs: Keyword arguments passed to :func:`fit`, defaults to None
        :type fit_kwargs: Optional[dict], optional
        :param sample_kwargs: Keyword arguments passed to :func:`sample`, defaults to None
        :type sample_kwargs: Optional[dict], optional
        :param `start_spread`: Keys are parameter names and values are range, defaults to None
        :type start_spread: Optional[dict[str, Iterable[float]]], optional
        :param smc: If True, sample all models using SMC, defaults to False
        :type smc: bool, optional
        :param approx: If True, approximate all models using VI, defaults to True
        :type approx: bool, optional
        """
        if self.verbose:
            print(f"Null hypothesis BIC = {self.models[1].null_bic():.3e}")
        if fit_kwargs is None:
            fit_kwargs = {}
        if sample_kwargs is None:
            sample_kwargs = {}

        if start_spread is not None and "start" not in fit_kwargs:
            fit_kwargs["start"] = {}
        if start_spread is not None and "init_kwargs" not in sample_kwargs:
            sample_kwargs["init_kwargs"] = {}
        if start_spread is not None and "start" not in sample_kwargs["init_kwargs"]:
            sample_kwargs["init_kwargs"]["start"] = {}

        stop = False
        for n_cloud in self.n_clouds:
            if approx:
                # fit with VI
                if self.verbose:
                    print(f"Approximating n_cloud = {n_cloud} posterior...")

                if start_spread is not None:
                    for key, value in start_spread.items():
                        fit_kwargs["start"][key] = np.linspace(
                            value[0], value[1], n_cloud
                        )

                self.models[n_cloud].fit(**fit_kwargs)
                if self.verbose:
                    bic = self.models[n_cloud].bic(chain=[0])
                    print(f"n_cloud = {n_cloud} BIC = {bic:.3e}")
                    print()

            else:
                if self.verbose:
                    print(f"Sampling n_cloud = {n_cloud} posterior...")
                if smc:
                    # sample with SMC
                    self.models[n_cloud].sample_smc(**sample_kwargs)
                else:
                    if start_spread is not None:
                        for key, value in start_spread.items():
                            sample_kwargs["init_kwargs"]["start"][key] = np.linspace(
                                value[0], value[1], n_cloud
                            )

                    # sample with MCMC
                    self.models[n_cloud].sample(**sample_kwargs)
                self.models[n_cloud].solve(kl_div_threshold=kl_div_threshold)
                if self.verbose:
                    for solution in self.models[n_cloud].solutions:
                        print(
                            f"n_cloud = {n_cloud} "
                            + f"solution = {solution} "
                            + f"BIC = {self.models[n_cloud].bic(solution=solution):.3e}"
                        )
                    print()

            # check stopping criteria
            if self._check_stop(n_cloud, bic_threshold=bic_threshold):
                if self.verbose:
                    print("Stopping criteria met.")
                # stop if we met criteria twice
                if stop:
                    if self.verbose:
                        print("Stopping early.")
                    break
                stop = True
            else:
                stop = False

        # get best model
        bics = self.bics
        n_clouds = list(bics.keys())
        model_bics = list(bics.values())
        best_idx = np.where(model_bics < (np.nanmin(model_bics) + bic_threshold))[0]
        if len(best_idx) == 0:
            if self.verbose:  # pragma: no cover
                print("No good models found!")
        else:
            best_n_clouds = n_clouds[best_idx[0]]
            self.best_model = self.models[best_n_clouds]

            if approx:
                # sample best
                if self.verbose:
                    print(
                        f"Sampling best model (n_cloud = {self.best_model.n_clouds})..."
                    )
                if smc:
                    self.best_model.sample_smc(**sample_kwargs)
                else:
                    if start_spread is not None:
                        for key, value in start_spread.items():
                            sample_kwargs["init_kwargs"]["start"][key] = np.linspace(
                                value[0], value[1], best_n_clouds
                            )

                    self.best_model.sample(**sample_kwargs)
                self.best_model.solve(kl_div_threshold=kl_div_threshold)



"""
test_gauss_models.py - Unit tests for GaussLine models and
bayes_spec sampling functions

Copyright(C) 2024 by
Trey V. Wenger; tvwenger@gmail.com
This code is licensed under MIT license (see LICENSE for details)
"""

import pytest

import numpy as np

import arviz as az
import graphviz

from bayes_spec import SpecData
from bayes_spec.models import GaussModel, GaussNoiseModel

import jax
import numpyro

jax.config.update("jax_platform_name", "cpu")
numpyro.set_platform("cpu")
numpyro.set_host_device_count(4)

_RNG = np.random.RandomState(seed=1234)

noise = 1.0
spectral = np.linspace(-100.0, 100.0, 1000)
dummy_brightness = noise * _RNG.randn(1000)
params = {
    "line_area": [1000.0],
    "fwhm": [25.0],
    "velocity": [10.0],
    "baseline_observation_norm": [0.0],
}

# Simulate single-component model
dummy_data = {"observation": SpecData(spectral, dummy_brightness, noise)}
_MODEL = GaussModel(dummy_data, 1, baseline_degree=0, seed=1234, verbose=True)
_MODEL.add_priors()
_MODEL.add_likelihood()
sim_brightness = _MODEL.model["observation"].eval(params)
_DATA = {"observation": SpecData(spectral, sim_brightness, noise)}
_MODEL = GaussModel(_DATA, 1, baseline_degree=0, seed=1234, verbose=True)
_MODEL.add_priors()
_MODEL.add_likelihood()


def test_gauss_model():
    assert isinstance(_MODEL.graph(), graphviz.sources.Source)
    assert isinstance(_MODEL.sample_prior_predictive(), az.InferenceData)
    with pytest.raises(ValueError):
        _MODEL.sample_posterior_predictive()


def test_gauss_model_nan():
    # Model with NaN data
    data = {"observation": SpecData(spectral, sim_brightness * np.nan, noise)}
    model = GaussModel(data, 1, baseline_degree=0, seed=1234, verbose=True)
    model.add_priors()
    model.add_likelihood()
    with pytest.raises(ValueError):
        model._validate()


def test_gauss_model_vi():
    # Fit single-component model with VI
    model = GaussModel(_DATA, 1, baseline_degree=0, seed=1234, verbose=True)
    model.add_priors(prior_baseline_coeffs=[1.0])
    model.add_likelihood()
    model.fit(rel_tolerance=0.01, abs_tolerance=0.1, learning_rate=1e-2)


def test_gauss_model_sample():
    # Sample single-component model
    model = GaussModel(_DATA, 1, baseline_degree=0, seed=1234, verbose=True)
    model.add_priors(prior_baseline_coeffs=[1.0])
    model.add_likelihood()
    model.sample(
        tune=1000,
        draws=1000,
        chains=4,
        cores=4,
        init_kwargs={
            "rel_tolerance": 0.01,
            "abs_tolerance": 0.1,
            "learning_rate": 1e-2,
            "start": {"velocity_norm": [-3.0]},
        },
    )
    model.solve()
    assert model.unique_solution
    assert model.bic() < model.null_bic()
    assert model.bic(chain=0) < model.null_bic()
    assert isinstance(model.sample_posterior_predictive(), az.InferenceData)
    assert isinstance(model.sample_posterior_predictive(solution=0), az.InferenceData)


def test_gauss_model_ordered_sample():
    # Sample single-component ordered model
    model = GaussModel(_DATA, 1, baseline_degree=0, seed=1234, verbose=True)
    model.add_priors(ordered=True)
    model.add_likelihood()
    model.sample(
        n_init=1000,
        tune=100,
        draws=100,
        chains=2,
        cores=2,
        init_kwargs={
            "rel_tolerance": 0.01,
            "abs_tolerance": 0.1,
            "learning_rate": 1e-2,
        },
    )
    model.solve()


def test_gauss_model_sample_auto():
    # Test "auto" initialization strategy
    _MODEL.sample(init="auto", n_init=1000, tune=100, draws=100, chains=2, cores=2)


def test_gauss_model_sample_smc():
    # Sample with SMC
    _MODEL.sample_smc(chains=2, draws=100)


def test_gauss_model_nutpie():
    # Sample with nutpie
    _MODEL.sample(nuts_sampler="nutpie", tune=100, draws=100)


def test_gauss_model_numpyro():
    # Sample with numpyro
    _MODEL.sample(nuts_sampler="numpyro", tune=100, draws=100)


def test_gauss_model_blackjax():
    _MODEL.sample(nuts_sampler="blackjax", tune=100, draws=100)


def test_gauss_model_prior_shape():
    # Regression tests for https://github.com/tvwenger/bayes_spec/issues/43
    model = GaussModel(_DATA, 1, baseline_degree=0, seed=1234, verbose=True)
    with pytest.raises(ValueError):
        model.add_priors(prior_line_area=[1.0, 1.0])
    with pytest.raises(ValueError):
        model.add_priors(prior_fwhm=[1.0, 1.0])
    with pytest.raises(ValueError):
        model.add_priors(prior_velocity=10.0)
    with pytest.raises(ValueError):
        model.add_priors(prior_velocity=[10.0])
    with pytest.raises(ValueError):
        model.add_priors(prior_baseline_coeffs=0.0)
    with pytest.raises(ValueError):
        model.add_priors(prior_baseline_coeffs={"observation": [1.0, 2.0, 3.0]})


def test_gauss_noise_model():
    # Fit single-component model with VI
    model = GaussNoiseModel(_DATA, 1, baseline_degree=0, seed=1234, verbose=True)
    model.add_priors(prior_baseline_coeffs=[1.0])
    model.add_likelihood()
    model.fit(rel_tolerance=0.01, abs_tolerance=0.1, learning_rate=1e-2)


def test_gauss_noise_model_prior_shape():
    # Regression tests for https://github.com/tvwenger/bayes_spec/issues/43
    model = GaussNoiseModel(_DATA, 1, baseline_degree=0, seed=1234, verbose=True)
    with pytest.raises(ValueError):
        model.add_priors(prior_rms=[1.0, 1.0])
